/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file mkdir.c
 * Implementation of the s3 DSI mkdir
 */
#include "globus_gridftp_server_s3_base.h"

static
void *
globus_l_s3_mkdir_thread(void *arg);

static
void
globus_l_s3_mkdir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_arg);

static
S3Status
globus_l_s3_mkdir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
int
globus_l_s3_mkdir_data_callback(
    int                                 bufferSize,
    char                               *buffer,
    void                               *callback_arg);

typedef struct
{
    /** GridFTP mkdir operation */
    globus_gfs_operation_t              op;
    /** Bucket Context derived from the session info and the mkdir pathname */
    S3BucketContext                     bucket_context;
    /** pointer into pathname that contains the bucket name */
    char                               *bucket_name;
    /** pointer into pathname that contains the key name */
    char                               *key;
    /** number of retries attempted */
    int                                 retries;
    /** S3's status of the operation */
    S3Status                            status;
    /** Result of the operation */
    globus_result_t                     result;
    /** Copy of the pathname string split into bucket_name and key */
    char                                pathname[];
}
globus_l_s3_mkdir_t;

/**
 * @brief mkdir
 * @details
 * For s3, there are not really any directories, so we just ensure that
 * an object called pathname/ exists so we can stat it later.
 */
void
globus_i_s3_mkdir(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_thread_t                     t;
    globus_l_s3_mkdir_t                *mkdir_state = NULL;
    int                                 rc;
    globus_result_t                     result;

    GlobusS3Enter();

    GlobusS3Info("mkdir: op=%s pathname=%s\n",
            (void *) op,
            cmd_info->pathname);

    assert(s3_handle != NULL);

    mkdir_state = malloc(sizeof(globus_l_s3_mkdir_t) + strlen(cmd_info->pathname) + 2);
    if (mkdir_state == NULL)
    {
        result = GlobusGFSErrorMemory("mkdir_state");

        goto failure;
    }

    *mkdir_state = (globus_l_s3_mkdir_t)
    {
        .op                          = op
    };

    strcpy(mkdir_state->pathname, cmd_info->pathname);
    globus_i_s3_path_parse(mkdir_state->pathname, &mkdir_state->bucket_name, &mkdir_state->key);

    mkdir_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, mkdir_state->bucket_name);

    if (mkdir_state->key)
    {
        strcat(mkdir_state->key, "/");
    }

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_mkdir_thread,
            mkdir_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");
failure:
        GlobusS3Info("finished_command: op=%p result=%#x path=%s\n",
                    (void *) op,
                    (unsigned int) result,
                    cmd_info->pathname);
        globus_gridftp_server_finished_command(op, result, NULL);
        free(mkdir_state);
    }

    GlobusS3Exit();

    return;
}
/* globus_i_s3_mkdir() */

static
void *
globus_l_s3_mkdir_thread(
    void                               *user_arg)
{
    globus_l_s3_mkdir_t                *mkdir_state = user_arg;
    S3PutProperties                     props = 
    {
        .contentType                  = GLOBUS_S3_DIR_PLACEHOLDER_TYPE
    };
    globus_result_t                     result = GLOBUS_SUCCESS;
    S3ResponseHandler                   response_handler =
    {
        .propertiesCallback           = globus_l_s3_mkdir_properties,
        .completeCallback             = globus_l_s3_mkdir_complete
    };
    S3PutObjectHandler                  putObjectHandler =
    {
        .responseHandler              = response_handler,
        .putObjectDataCallback        = globus_l_s3_mkdir_data_callback
    };

    GlobusS3Enter();

    do
    {
        mkdir_state->status = S3StatusOK;
        mkdir_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(mkdir_state->retries);

        if (mkdir_state->key)
        {
            GlobusS3Info("REST: op=%p S3_put_object(bucket=%s, key=%s) try=%d\n",
                        (void *) mkdir_state->op,
                        mkdir_state->bucket_name,
                        mkdir_state->key,
                        mkdir_state->retries);

            S3_put_object(
                    &mkdir_state->bucket_context,
                    mkdir_state->key,
                    0,
                    &props,
                    NULL,
                    &putObjectHandler,
                    mkdir_state);

            GlobusS3DebugOpStatus(S3_put_object, mkdir_state);
        }
        else
        {
            GlobusS3Info("REST: op=%p S3_create_bucket(bucket=%s) try=%d\n",
                        (void *) mkdir_state->op,
                        mkdir_state->bucket_name,
                        mkdir_state->retries);

            S3_create_bucket(
                mkdir_state->bucket_context.protocol,
                mkdir_state->bucket_context.accessKeyId,
                mkdir_state->bucket_context.secretAccessKey,
                mkdir_state->bucket_context.securityToken,
                NULL,
                mkdir_state->bucket_name,
                0,
                NULL,
                NULL,
                &putObjectHandler.responseHandler,
                mkdir_state);

            GlobusS3DebugOpStatus(S3_create_bucket, mkdir_state);
        }
    }
    while (S3_status_is_retryable(mkdir_state->status)
            && (mkdir_state->retries++) < globus_i_s3_max_retries);

    result = mkdir_state->result;

    GlobusS3Info("finished_command: op=%p result=%#x path=%s\n",
                (void *) mkdir_state->op,
                (unsigned int) mkdir_state->result,
                mkdir_state->pathname);
    globus_gridftp_server_finished_command(mkdir_state->op, result, NULL);

    free(mkdir_state);

    return NULL;
}
/* globus_i_s3_mkdir_thread() */

static
S3Status
globus_l_s3_mkdir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_mkdir_properties() */

static
int
globus_l_s3_mkdir_data_callback(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_arg)
{
    GlobusS3Enter();

    /* Return 0 to indicate no content to send */

    GlobusS3Exit();

    return 0;
}
/* globus_l_s3_mkdir_data_callback() */

static
void
globus_l_s3_mkdir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_arg)
{
    globus_l_s3_mkdir_t                *mkdir_state = callback_arg;

    GlobusS3Enter();

    if (status != S3StatusOK && mkdir_state->result == GLOBUS_SUCCESS)
    {
        mkdir_state->status = status;
        mkdir_state->result = GlobusGFSErrorS3(
                    status,
                    mkdir_state->key,
                    error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_mkdir_complete() */
// vim: ft=c:
